from __future__ import annotations

import dataclasses
import typing as typ

from numbers import Number

import jijmodeling as jm
import numpy as np

from jijzept.config.path_type import PATH_TYPE
from jijzept.entity.schema import SolverType
from jijzept.response import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    check_kwargs_against_dataclass,
    merge_params_and_kwargs,
    sample_model,
)
from jijzept.sampler.sample_hubo import sample_hubo
from jijzept.sampler.sample_qubo import sample_qubo
from jijzept.sampler.utils import _set_needs_square

T = typ.TypeVar("T")


class QIOParameters:
    pass


P = typ.TypeVar("P")


class JijQIOBaseSampler(JijZeptBaseSampler, typ.Generic[P]):
    parameter_cls: typ.Type[QIOParameters]

    def __init__(
        self,
        token: str | None = None,
        url: str | None = None,
        proxy: str | None = None,   
        config: PATH_TYPE | None = None,
        config_env: str = "default",
    ) -> None:
        super().__init__(token, url, proxy, config, config_env)

    def sample_model(
        self,
        model: jm.Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], int | float]] = {},
        needs_square_dict: dict[str, bool] | None = None,
        search: bool = False,
        num_search: int = 15,
        algorithm: str | None = None,
        parameters: P | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of Microsoft QIO Simulated Annealing solver.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            model (jijmodeling.Problem): Mathematical model is constracted by JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            needs_square_dict (Dict[str, bool], optional): If `True`, the corresponding constraint is squared when added as a penalty to the QUBO. When the constraint label is not added to the 'needs_square_dict', it defaults to `True` for linear constraints and `False` for non-linear constraints.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            parameters (P | None, optional): parameters for QIO. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 60 (one minute) will be set. Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist,　the value of `**kwargs` takes precedence.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            import jijmodeling as jm
            n = jm.Placeholder('n')
            x = jm.Binary('x', shape=n)
            d = jm.Placeholder('d', shape=n)
            i = jm.Element("i", n)
            problem = jm.Problem('problem')
            problem += jm.Sum(i, d[i] * x[i])
            sampler = jz.JijQIOSASampler(config='config.toml')
            response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
            ```
        """
        model = _set_needs_square(model, needs_square_dict)

        check_kwargs_against_dataclass(kwargs, self.parameter_cls)
        param_dict = merge_params_and_kwargs(parameters, kwargs, self.parameter_cls)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        para_search_params = ParameterSearchParameters(
            multipliers=multipliers,
            mul_search=search,
            num_search=num_search,
            algorithm=algorithm,
        )

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=model,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set

    def sample_qubo(
        self,
        Q: dict,
        parameters: P | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Microsoft QIO solver is performed for binary quadratic models.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            Q (Dict): The linear or quadratic terms.
            parameters (P | None, optional): parameters for QIO. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 60 (one minute) will be set. Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist, the value of `**kwargs` takes precedence.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            sampler = jz.JijQIOSASampler(config='config.toml')
            response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
            ```
        """
        check_kwargs_against_dataclass(kwargs, self.parameter_cls)
        param_dict = merge_params_and_kwargs(parameters, kwargs, self.parameter_cls)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_qubo(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            qubo=Q,
            constant=0,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )

        return sample_set

    def sample_hubo(
        self,
        J: dict,
        parameters: P | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Microsoft QIO simulated annealing solver is performed for binary polynomial models.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            J (Dict): Polynomial interactions.
            parameters (P | None, optional): parameters for QIO. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If None, 60 (one minute) will be set. Please note that this argument is for the jijzept timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist, the value of `**kwargs` takes precedence.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            sampler = jz.JijQIOSASampler(config='config.toml')
            response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
            ```
        """
        check_kwargs_against_dataclass(kwargs, self.parameter_cls)
        param_dict = merge_params_and_kwargs(parameters, kwargs, self.parameter_cls)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_hubo(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            polynomial=J,
            constant=J.get((), 0.0),
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set
